/*=============================================================================
 * parser for CSP instances represented in XML format
 * 
 * Copyright (c) 2008 Olivier ROUSSEL (olivier.roussel <at> cril.univ-artois.fr)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *=============================================================================
 */
#ifndef _CSPNormalizerCallback_hh_
#define _CSPNormalizerCallback_hh_

#include <vector>
#include <algorithm>
#include "CSPParserCallback.hh"

//#define PrettyPrint

/**
 * TODO:
 * - sort the tuples of a relation
 * - check the arity of a relation
 * - recompute the scope of predicates and global constraints
 * - recompute and normalize domains
 * - remove domains or variables which are not referenced
 */

namespace CSPXMLParser
{
  using namespace std;

  class NormalizerASTFactory: public DefaultASTFactory
  {
  public:
    class NormalizerASTVar: public ASTVar
    {
    public:
      NormalizerASTVar(const string &name, int id=-1): ASTVar(name,id) {}

      virtual void prefixExpression(ostream &s) const
      {
	if (inFormalParameters)
	  s << "X" << id;
	else
	  s << "V" << id;
      }

      static bool inFormalParameters;
    };

    class NormalizerASTSymb: public ASTSymb
    {
    public:
      NormalizerASTSymb(NodeType c): ASTSymb(c) {}

      virtual void prefixExpression(ostream &s) const
      {
	switch(cst)
	{
	case SYMB_NIL:
	  s << "<nil/>";
	  break;
	case SYMB_EQ:
	  s << "<eq/>";
	  break;
	case SYMB_NE:
	  s << "<ne/>";
	  break;
	case SYMB_GE:
	  s << "<ge/>";
	  break;
	case SYMB_GT:
	  s << "<gt/>";
	  break;
	case SYMB_LE:
	  s << "<le/>";
	  break;
	case SYMB_LT:
	  s << "<lt/>";
	  break;
	default:
	  throw runtime_error("unimplemented");
	}
      }
    };

    class NormalizerASTDict: public ASTDict
    {
    public:
      // ??? temporary solution
      static vector<string> *convOrder; // conventional order to use (if any)

      virtual void prefixExpression(ostream &s) const
      {
	s << "{";

	if (convOrder)
	{
	  for(vector<string>::iterator it=convOrder->begin();
	      it!=convOrder->end();++it)
	    if (hasKey(*it))
	    {
	      s << ' '; 
	      (*this)[*it].prefixExpression(s);
	    }
	    else
	      s << " <nil/>";
	}
	else
	  for(map<string,AST *>::const_iterator it=dict.begin();
	      it!=dict.end();++it)
	  {
	    s << " /" << (*it).first << ' ';
	    (*it).second->prefixExpression(s);
	  }

	s << " }";
      }
    };

    inline static ASTVar *mkVar(const string &name, int id=-1) 
    {
      return new NormalizerASTVar(name,id);
    }

    inline static ASTSymb *mkSymb(NodeType c) 
    {
      return new NormalizerASTSymb(c);
    }

    inline static ASTDict *mkDict() 
    {
      return new NormalizerASTDict();
    }
  };

  bool NormalizerASTFactory::NormalizerASTVar::inFormalParameters=false;
  vector<string> *NormalizerASTFactory::NormalizerASTDict::convOrder=NULL;

  class CSPNormalizerCallback : public CSPParserCallback
  {
  private:
    bool firstDomainValue;
    bool firstTuple;
    string constraintReference;
    bool noParms;

    bool pass1;

    bool isWCSP;

    int nbDomain,nbVar,nbRel,nbPred,nbConstr,maxConstraintArity;

    // ??? temporary solution
    // conventional order for some global constraints
    vector<string> cumulative_dictOrder,weightedSum_dictOrder;

  public:
    CSPNormalizerCallback()
    {
      nbDomain=0;
      nbVar=0;
      nbRel=0;
      nbPred=0;
      nbConstr=0;
      maxConstraintArity=0;

      cumulative_dictOrder.push_back("origin");
      cumulative_dictOrder.push_back("duration");
      cumulative_dictOrder.push_back("end");
      cumulative_dictOrder.push_back("height");

      weightedSum_dictOrder.push_back("coef");
      weightedSum_dictOrder.push_back("var");
    }

    /**
     * the first pass only counts
     */
    void firstPass()
    {
      pass1=true;
    }

    void secondPass()
    {
      pass1=false;
    }

    virtual void beginInstance(const string & name)
    {
      if (pass1)
	return;

      cout << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
	   << "<instance>" << endl;

      cout << "<presentation maxConstraintArity=\"" 
	   << maxConstraintArity << "\" format=\"XCSP 2.1\"";
    }

    virtual void instanceType(InstanceType instanceType) 
    {
      if (pass1)
	return;

      isWCSP=false;

      switch(instanceType)
      {
      case CSP_INSTANCE:
	cout << "/>" << endl;
	break;
      case WCSP_INSTANCE:
	isWCSP=true;
	cout << " type=\"WCSP\"/>" << endl;
	break;
      default:
	throw runtime_error("unsupported type of instance");
	break;
      }
    }

    virtual void beginDomainsSection(int n) 
    {  
      if (pass1)
	return;

      cout << "<domains nbDomains=\"" 
	   << nbDomain << "\">" <<endl;
    }
  
    virtual void beginDomain(const string & name, int idDomain, int nbValue) 
    {
      if (pass1)
      {
	++nbDomain;
	return;
      }

      cout << " <domain name=\"D" << idDomain 
	   << "\" nbValues=\"" << nbValue << "\">";
      firstDomainValue=true;
    }

    virtual void addDomainValue(int v) 
    {
      if (pass1)
	return;

      if (!firstDomainValue)
	cout << ' ';
      firstDomainValue=false;

      cout << v;
    }

    virtual void addDomainValue(int first,int last) 
    {
      if (pass1)
	return;

      if (!firstDomainValue)
	cout << ' ';
      firstDomainValue=false;

      cout << first << ".." << last;
    }

    virtual void endDomain() 
    {
      if (pass1)
	return;

      cout << "</domain>" <<endl;
    }

    /**
     * end the definition of all domains
     */
    virtual void endDomainsSection() 
    {
      if (pass1)
	return;

      cout << "</domains>" <<endl;
    }


    virtual void beginVariablesSection(int nbVariables) 
    {
      if (pass1)
	return;

      cout << "<variables nbVariables=\""
	   << nbVar << "\">" <<endl;
    }
  
    virtual void addVariable(const string & name, int idVar,
			     const string & domain, int idDomain) 
    {
      if (pass1)
      {
	++nbVar;
	return;
      }

      cout << " <variable name=\"V" << idVar
	   << "\" domain=\"D" << idDomain 
	   << "\"/>" <<endl;
    }

    virtual void endVariablesSection() 
    {
      if (pass1)
	return;

      cout << "</variables>" <<endl;
    }


    virtual void beginRelationsSection(int nbRelations) 
    {
      if (pass1)
	return;

      cout << "<relations nbRelations=\"" << nbRel
	   << "\">" <<endl;
    }

  
    virtual void beginRelation(const string & name, int idRel,
			       int arity, int nbTuples, RelType relType) 
    {
      curIdRel=idRel;

      if (pass1)
      {
	relInfo.push_back(RelationInfo());

	return;
      }

      if (nbTuples!=relInfo[idRel].nbTuples)
	cerr << "WARNING: fixing the number of tuples in " << name << endl;

      if (relInfo[idRel].nbTuples==0)
      {
	/*
	if (relType==REL_SUPPORT)
	  throw runtime_error("Trivially UNSAT");
	*/
	cerr << "WARNING: ignoring trivial relation" << endl;

	return;
      }

      cout << "<relation name=\"R" << relInfo[idRel].newId 
	   << "\" arity=\"" << arity << "\" nbTuples=\""
	   << relInfo[idRel].nbTuples << "\" semantics=\"";

      switch(relType)
      {
      case REL_SUPPORT:
	cout << "supports";
	break;
      case REL_CONFLICT:
	cout << "conflicts";
	break;
      case REL_SOFT:
	cout << "soft";
	break;
      default:
	throw runtime_error("unknown relation type");
      }

      firstTuple=true;
      lastCost=0;

      if (relType!=REL_SOFT)
	cout << "\">\n";
    }

    virtual void relationDefaultCost(const string & name, int idRel,
			     int defaultCost)
    {
      if (pass1)
	return;

      cout << "\" defaultCost=\"" << defaultCost << "\">\n";      
    }

    virtual void addRelationTuple(int arity, int tuple[]) 
    {      
      if (pass1)
      {
	relInfo.back().nbTuples++;
	return;
      }

      if (!firstTuple)
	cout << '|';

      firstTuple=false;

      cout << tuple[0];
      for(int i=1;i<arity;++i)
	cout << ' ' << tuple[i];
    }

    virtual void addRelationTuple(int arity, int tuple[], int cost) 
    {      
      if (pass1)
      {
	relInfo.back().nbTuples++;
	return;
      }

      if (!firstTuple)
	cout << '|';

      firstTuple=false;

      if (cost!=lastCost)
      {
	cout << cost << ":";
	lastCost=cost;
      }

      cout  << tuple[0];
      for(int i=1;i<arity;++i)
	cout << ' ' << tuple[i];
    }

    virtual void endRelation() 
    {
      if (relInfo[curIdRel].nbTuples==0)
	return;

      if (pass1)
      {
	relInfo.back().newId=nbRel;
	++nbRel;

	return;
      }


      cout << "\n</relation>" <<endl;
    }

    virtual void endRelationsSection() 
    {
      if (pass1)
	return;

      cout << "</relations>" <<endl;
    }

    virtual void beginPredicatesSection(int nbPredicates) 
    {
      if (pass1)
	return;

      cout << "<predicates nbPredicates=\"" 
	   << nbPred << "\">" << endl;
    }
  
    virtual void beginPredicate(const string & name, int idPred) 
    {
      if (pass1)
      {
	++nbPred;
	return;
      }

      cout << "<predicate name=\"P" << idPred 
	   << "\">\n" 
	   << " <parameters>";

      idFormParm=0;
      NormalizerASTFactory::NormalizerASTVar::inFormalParameters=true;
    }

    virtual void addFormalParameter(int pos, const string & name, 
				    const string & type) 
    {
      if (pass1)
	return;

      if (idFormParm!=0)
	cout << " ";

      cout << type << " X" << idFormParm++;
    }

    virtual void predicateExpression(AST *tree) 
    {
      if (pass1)
	return;

      cout << "   predicate definition (AST) = ";
      tree->prefixExpression(cout);
      cout << endl;
    }

    virtual void predicateExpression(const string &expr) 
    {
      if (pass1)
	return;

      NormalizerASTFactory::NormalizerASTVar::inFormalParameters=false;
      cout << "</parameters>\n"
	   << " <expression>\n"
	   << "  <functional>"
#ifdef PrettyPrint
	   << "\n   "
#endif
	   << expr
#ifdef PrettyPrint
	   << "\n  "
#endif
	   << "</functional>\n"
	   << " </expression>\n";
    }

    virtual void endPredicate() 
    {
      if (pass1)
	return;

      cout << "</predicate>" << endl;
    }

    virtual void endPredicatesSection() 
    {
      if (pass1)
	return;

      cout << "</predicates>" << endl;
    }

    virtual void beginConstraintsSection(int nbConstraints) 
    {
      if (pass1)
	return;

      cout << "<constraints nbConstraints=\""
	   << nbConstr << "\"";

      if (!isWCSP)
	cout << ">" <<endl;
    }

    virtual void constraintsMaximalCost(int maximalCost)
    {
      if (pass1)
	return;

      cout << " maximalCost=\"" << maximalCost << "\">\n";      
    }
  
    virtual void beginConstraint(const string & name, int idConstr,
				 int arity, 
				 const string & reference, 
				 CSPDefinitionType type, int id,
				 const ASTList &scope)
    {
      // ignore trivial relations
      if (type==RelationType && relInfo[id].nbTuples==0)
      {
	ignoreConstraint=true;
	return;
      }
      else
	ignoreConstraint=false;

      if (!pass1)
	return;

      curConstrDef.str("");

      ++nbConstr;

      // ??? move 
      if (scope.size()>maxConstraintArity)
	maxConstraintArity=scope.size();

      curConstrDef << " arity=\"" << scope.size()
		   << "\" scope=\"";


      curScope.clear();
      for(int i=0;i<scope.size();++i)
	curScope.push_back(scope[i].getVarId());

      sort(curScope.begin(),curScope.end());

      if (type==RelationType)
      {
	// scope order is significant
	if (scope.size())
	  curConstrDef << "V" << scope[0].getVarId();
	
	for(int i=1;i<scope.size();++i)
	  curConstrDef << " V" << scope[i].getVarId();
      }
      else
      {
	// scope order is not significant
	if (curScope.size())
	  curConstrDef << "V" << curScope[0];
	
	for(int i=1;i<curScope.size();++i)
	  curConstrDef << " V" << curScope[i];
      }

      noParms=false;
      allDifferent=reference=="global:alldifferent";

      curConstrDef << "\" reference=\"";
      switch(type)
      {
      case RelationType:
	curConstrDef << "R" << relInfo[id].newId;
	noParms=true;
	break;
      case PredicateType:
	curConstrDef << "P" << id;
	break;
      case GlobalConstraintType:
	curConstrDef << reference;

	NormalizerASTFactory::NormalizerASTDict::convOrder=NULL;
	if (reference=="global:cumulative")
	  NormalizerASTFactory::NormalizerASTDict::convOrder=&cumulative_dictOrder;
	else
	  if (reference=="global:weightedsum")
	    NormalizerASTFactory::NormalizerASTDict::convOrder=&weightedSum_dictOrder;

	break;
      default:
	throw runtime_error("unexpected type of constraint");
      }

      if (noParms)
	curConstrDef << "\"/>" << endl;
      else
	curConstrDef << "\">" << endl;

      constraintReference=reference;
    }

    virtual void constraintParameters(const ASTList &args)
    {
      if (!pass1)
	return;

      if (noParms)
	return;

      curConstrDef << "  <parameters>"
#ifdef PrettyPrint
		   << "\n  "
#endif
	;

      // cope with the deprecated syntax which has no parameters for
      // allDifferent (in this case, the parser gives the scope as
      // parameter)
      if (allDifferent && (args.size()==0 || !args[0].isList()))
	curConstrDef << " [";

      args[0].prefixExpression(curConstrDef);
      for(int i=1;i<args.size();++i)
      {
	curConstrDef << ' ';
	args[i].prefixExpression(curConstrDef);
      }

      if (allDifferent && (args.size()==0 || !args[0].isList()))
	curConstrDef << " ]";

      curConstrDef 
#ifdef PrettyPrint	
	           << "\n  " 
#endif
		   << "</parameters>\n";
    } 


    virtual void endConstraint() 
    {
      if (!pass1 || ignoreConstraint)
	return;

      if (!noParms)
	curConstrDef << " </constraint>" << endl;

      constraints.insert(make_pair(curScope,curConstrDef.str()));
    }

    /**
     * end the definition of all constraints
     */
    virtual void endConstraintsSection() 
    {
      if (pass1)
	return;

      int cid=0;

      for(multimap<vector<int>,string>::iterator it=constraints.begin();
	  it!=constraints.end();++it)
      {
	cout << " <constraint name=\"C" << cid++ << "\"" << (*it).second;
      }

      cout << "</constraints>" <<endl;
    }

    /********************************************************************/


    /**
     * signal the end of parsing
     */
    virtual void endInstance() 
    {
      if (pass1)
	return;

      cout << "</instance>" <<endl;
    }


  private:
    class ScopeOrder
    {
    public:
      bool operator() (const vector<int> &a, const vector<int> &b)
      {
	if (a.size()<b.size())
	  return true;
	else
	  if (a.size()>b.size())
	    return false;
	  else
	    return a<b;
      }
    };

    bool allDifferent;
    int lastCost;
    int idFormParm;

    vector<int> curScope;
    ostringstream curConstrDef;

    // maps scope to constraints
    multimap<vector<int>,string,ScopeOrder> constraints;

    // information on relations
    struct RelationInfo
    {
      int newId; // new identifier after removal of trivial constraints
      int nbTuples;

      RelationInfo()
      {
	newId=-1;
	nbTuples=0;
      }
    };

    int curIdRel;
    vector<RelationInfo> relInfo;

    bool ignoreConstraint;
  };

}

#endif
